(** Proof of type soundness for the lambda calculus with references. 
Proof based on the development in Pierce, "Types and Programming Languages" *)

(** copyright Mulhern *)

(** General notion of a map. *)

Require Import List.
Require Peano_dec.
Require Omega.

Require Import util.

(** We define a map as a partial function from labels to values. 
We define here only the properties on a map that remains constant in size,
i.e., we can lookup and update but we cannot insert. *)
Module Map (VAL : OPTION).

  Module Opt := VAL.
  
  (** [V] is the type of the actual value stored. *)
  Definition V := VAL.V.A.
  
  (** The implementation of the map is a list of pairs. *)
  Definition map := list (nat * V).
  
  (** The domain of the map is the first element of every pair. *)
  Definition dom (s : map) := List.map (fst (B := V)) s.
  
  (** Here we prove that a map has a generic well-formedness property. *)
  Section WellFormed.
  
    (** A well-formed list does not repeat labels. *)
    Inductive well_formed : map -> Prop :=
      | NilWellFormed : well_formed nil 
      | ConsWellFormed : forall (s : map ) (n : nat) (b : V),
          well_formed s ->
            ~ In n (dom s) ->
              well_formed ((n,b)::s).

    Implicit Arguments ConsWellFormed [s n b].

    Derive Inversion_clear well_formed_cons_inv 
      with (forall (n : nat) (b : V) (s : map), well_formed ((n,b)::s)) Sort Prop.
    
    (** If a map is well formed then so is its tail. *)
    Lemma well_formed_inv : forall (m : map),
      well_formed m -> 
        match m with 
          | nil => True
          | hd::tl => ~ In (fst hd) (dom tl) /\ well_formed tl
        end
    .
    Proof.
      intros.
      induction m.
      auto.
      destruct a.
      inversion H using well_formed_cons_inv.
      auto.
    Qed.
    
    (** [well_formed_gt] is a stronger property than [well_formed].
    In this case, we require that each label is strictly greater than the
    preceding labels. *)
    Inductive well_formed_gt : map -> Prop :=
      | NilWellFormedGt : well_formed_gt nil
      | ConsWellFormedGt : forall (s : map) (n : nat) (b : V),
          well_formed_gt s ->
            (forall z : nat, In z (dom s) <-> n > z) ->
              well_formed_gt ((n,b)::s).

    Implicit Arguments ConsWellFormedGt [s n b].

    Derive Inversion_clear well_formed_gt_cons_inv 
      with (forall (n : nat) (b : V) (s : map), well_formed_gt ((n,b)::s)) Sort Prop.

    Lemma gt_stronger : forall (s : map),
      well_formed_gt s -> well_formed s.
    Proof.
      induction s.
      intros.
      constructor.
      intros.
      destruct a.
      inversion_clear H.
      constructor.
      apply IHs.
      assumption.
      red.
      intro.
      generalize (proj1 (H1 n) H).
      omega.
    Qed.

  End WellFormed.
  
  (** The internal implementation of lookup. *)
  Fixpoint lookup (s : map) (n : nat) {struct s } : Opt.A :=
    match s with
      | nil => None
      | h::t => match Peano_dec.eq_nat_dec (fst h) n with
          | left _ => Some (snd h)
          | right _ => lookup t n
         end
    end
  .
  
  (** [lookup] has the property that we'll find a value iff the 
  label is in the map.*)
  Lemma lookup_correct : forall (s : map) (n : nat),
    ((ex (fun b => Opt.equal (lookup s n) (Some b))) <-> In n (dom s)) /\
    (Opt.equal (lookup s n) (None) <-> ~ In n (dom s)).
    Proof.
      split.
      induction s.
      (** Base case *)
      simpl.
      split.
      intros.
      destruct H.
      inversion H.
      tauto.
      (** Induction step *)
      destruct a.
      simpl.
      induction (Peano_dec.eq_nat_dec n0 n).
      split.
      intros.
      tauto.
      intros.
      exists v.
      apply Opt.equal_refl.
      split.
      intros.
      right.
      apply (proj1 (IHs)).
      auto.
      tauto.
      induction s.
      (** Base case *)
      simpl.
      split.
      tauto.
      intros.
      apply Opt.equal_refl.
      (** Induction step *)
      destruct a.
      simpl.
      induction (Peano_dec.eq_nat_dec n0 n).
      split.
      intros.
      inversion_clear H.
      intros.
      red in H.
      generalize (H (or_introl _ a)).
      tauto.
      split.
      intros. 
      red.
      intros.
      destruct H0.
      contradiction.
      apply ((proj1 IHs) H H0).
      intros.
      apply (proj2 IHs).
      red.
      intros.
      auto.
  Qed.
  
  (** Two maps are equal iff all entries are equal. We cannot use =. *)
  Inductive equal : map  -> map  -> Prop :=
    | EqualNil : equal nil nil
    | EqualCons : forall (n n' : nat) (v v' : V) (m m' : map),
         n = n' -> VAL.V.equal v v' -> equal m m' -> equal ((n,v)::m) ((n',v')::m')
  .

  Derive Inversion_clear equal_cons_inv 
    with (forall (n n': nat) (v v' : V) (m m' : map), equal ((n,v)::m) ((n',v')::m') )Sort Prop.

  (** equivalence properties on maps are proved. *)
  Lemma equal_refl : forall (x : map), equal x x.
  Proof.
    intros.
    induction x.
    constructor.
    destruct a.
    constructor.
    reflexivity.
    apply VAL.V.equal_refl.
    assumption.
  Qed.

  Lemma equal_sym : forall (x y : map), equal x y -> equal y x.
  Proof.
    induction x; induction y; try solve [intros; inversion H].
    auto.
    intros.
    destruct a0; destruct a.
    inversion H using equal_cons_inv.
    intros.
    constructor.
    symmetry.
    assumption.
    apply VAL.V.equal_sym.
    assumption.
    apply IHx.
    assumption.
  Qed.

  Lemma equal_trans : forall (x y z : map), equal x y -> equal y z -> equal x z.
  Proof.
    induction x.
    intros.
    inversion H.
    subst y.
    assumption.
    induction y.
    intros.
    inversion H.
    induction z.
    intros.
    inversion H0.
    destruct a; destruct a0; destruct a1.
    intros.
    inversion H0 using equal_cons_inv.
    inversion H using equal_cons_inv.
    intros.
    constructor.
    transitivity n0; assumption.
    apply (VAL.V.equal_trans v v0 v1); assumption.
    eapply IHx.
    apply H3.
    apply H6.
  Qed.

  (** If two maps are equal then lookup on either map will yield the 
   same result *)
  Lemma lookup_equivalent : forall (s s' : map),
    equal s s' -> forall (n : nat), Opt.equal (lookup s n) (lookup s' n).
  Proof.
    induction s; induction s'; intros; try solve [inversion H].
    simpl.
    constructor.
    destruct a; destruct a0.
    simpl.
    inversion_clear H.
    induction (Peano_dec.eq_nat_dec n0 n);
    induction (Peano_dec.eq_nat_dec n1 n).
    constructor.
    assumption.
    subst n.
    symmetry in H0.
    contradiction.
    subst n.
    contradiction.
    apply IHs.
    assumption.
  Qed.

  (** update sets a new value iff the key is already in the map *)
  Fixpoint update (s : map) (n : nat) (newb : V) {struct s} : map :=
    match s with 
      | nil => nil
      | h::t => match Peano_dec.eq_nat_dec n (fst h) with
          | left _ => (fst h, newb)::t
          | right _ => h::update t n newb
         end
    end
  .

  (** updating does not change the keys *)
  Lemma update_preserves_dom : forall (m : map) (n : nat) (newb : V),
    dom(update m n newb) = dom m.
  Proof.
    induction m.
    simpl.
    auto.
    destruct a.
    simpl.
    intros.    
    induction (Peano_dec.eq_nat_dec n0 n).
    simpl.
    reflexivity.
    simpl.
    apply (f_equal (cons n)).
    apply IHm.
  Qed.
  
  (** If the key is not in the map then the map is unchanged *)
  Lemma update_preserves_map : forall (m : map) (n : nat) (newb :V),
    ~In n (dom m) -> equal m (update m n newb).
  Proof.
    induction m.
    intros.
    simpl.
    constructor.
    intros.
    destruct a.
    simpl.
    induction (Peano_dec.eq_nat_dec n n0).
    constructor.
    reflexivity.
    subst n.
    simpl in H.
    tauto.
    apply equal_refl.
    constructor.
    reflexivity.
    apply VAL.V.equal_refl.
    apply IHm.
    red.
    intro.
    apply H.
    simpl.
    auto.
  Qed.
  
  (** [update] will leave all values that do not match the key untouched,
         [lookup] will find what update just updated iff the key is in the map,
         [lookup] will find nothing iff the key is not in the map *)
  Lemma update_correct : forall (s : map) (n : nat) (newB : V),
    (forall (n' : nat), n' <> n -> Opt.equal (lookup (update s n newB) n') (lookup s n')) /\
    (In n (dom s) <-> Opt.equal (lookup (update s n newB) n) (Some newB)) /\
     (~In n (dom s) <-> Opt.equal (lookup (update s n newB) n) None).
  Proof.
    induction s.
    (** Base case *)
    simpl.
    split.
    intros.
    constructor.
    split.
    split.
    tauto.
    intros.
    inversion H. 
    split. 
    intros. 
    constructor. 
    tauto.
    (** Induction step. *)
    destruct a.
    simpl.
    intros.
    induction (Peano_dec.eq_nat_dec n0 n).
    subst n0.
    split.
    intros.
    induction (Peano_dec.eq_nat_dec n n').
    symmetry in a.
    contradiction.
    simpl. 
    induction (Peano_dec.eq_nat_dec n n').   
    contradiction.
    apply Opt.equal_refl.
    simpl.
    induction (Peano_dec.eq_nat_dec n n).
    split.
    split. 
    intros. 
    constructor. 
    apply Opt.V.equal_refl. 
    auto.
    split.
    intros.
    red in H.
    generalize (H (or_introl _ a)).
    tauto.
    intros.
    inversion H.
    tauto.
    intros.
    simpl.
    split.
    intros.
    induction (Peano_dec.eq_nat_dec n n').
    constructor.
    apply Opt.V.equal_refl.
    apply (proj1 (IHs n0 newB)).
    auto.
    induction (Peano_dec.eq_nat_dec n n0).
    symmetry in a.
    contradiction.
    split.
    split.
    intros. 
    destruct H. 
    contradiction.  
    apply (proj1 (proj1 (proj2 (IHs n0 newB)))). 
    assumption. 
    intros.
    right.
    apply (proj2 (proj1 (proj2 (IHs n0 newB)))). 
    assumption. 
    split. 
    intros. 
    apply (proj1 (proj2 (proj2 (IHs n0 newB)))). 
    tauto. 
    intros. 
    red.
    intros. 
    destruct H0.
    contradiction.
    generalize ((proj2 (proj2 (proj2 (IHs n0 newB)))) H).
    tauto.
  Qed.

  (** a well formed map is well formed after an update *)
  Lemma update_preserves_wf : forall (m : map) (n : nat) (newb : V),
    well_formed m -> well_formed (update m n newb).
  Proof.
    induction m.
    simpl.
    auto.
    destruct a.
    simpl.
    intros. 
    inversion_clear H.
    induction (Peano_dec.eq_nat_dec n0 n).
    subst n0.
    constructor.
    assumption.
    assumption.
    constructor.
    apply IHm.
    assumption.
    red.
    intro.
    rewrite (update_preserves_dom m n0 newb) in H.
    contradiction.
  Qed.

  (** a map with the stronger property well_formed_gt is well_formed_gt
      after an update *)
  Lemma update_preserves_wf_gt : forall (m : map) (n : nat) (newb : V),
    well_formed_gt m -> well_formed_gt (update m n newb).
  Proof.
    induction m.
    simpl.
    auto.
    destruct a.
    simpl.
    intros.
    inversion_clear H. 
    induction (Peano_dec.eq_nat_dec n0 n).
    subst n0.
    constructor.
    assumption.
    intros.
    apply H1.
    constructor.
    apply IHm.
    assumption.
    intros.
    rewrite (update_preserves_dom m n0 newb).
    apply (H1 z).
  Qed.   

End Map.
