(** Proof of type soundness for the lambda calculus with references. 
Proof based on the development in Pierce, "Types and Programming Languages" *)

(** copyright Mulhern *)

Require Import List.

Require Gt.
Require Compare_dec.

Require Import util.
Require Import store.

(** [label]s are natural numbers because there are an infinite number of 
natural numbers and they are all unique *)
Definition label := nat.

(** The rudimentary language contains only two types. *) 
Inductive ty : Set :=
  | TyRef : ty -> ty
  | TyUnit : ty
.

(** Make a module of [ty]---we want it to implement [VALUE] because we store 
types in the store typing. *) 
Module Ty <: VALUE.

  Definition A := ty. 
  Definition equal (t1 t2 : A) := eq t1 t2. 
  Lemma equal_refl : forall (t : A), equal t t.
  Proof.
  intros.
  unfold equal.
  reflexivity.
  Qed.
  Lemma equal_sym : forall (x y : A), equal x y -> equal y x.
  Proof.
  unfold equal.
  intros.
  symmetry.
  assumption.
  Qed.
  Lemma equal_trans : forall (x y z : A), equal x y -> equal y z -> equal x z.
  Proof.
  unfold equal.
  intros.
  subst x.
  subst z.
  reflexivity.
  Qed.

End Ty.

(** The [StoreTyping] module maps labels to types. **)

Module TyOpt := Option(Ty).

Module StoreTyping := Map_Wf_Ext (TyOpt).

(** We want to make references, dereference, and also do assignments. The
result of an assignment is unit. A location just wraps a label. *) 

Inductive term : Set :=
  | TmRef : term -> term
  | TmDeref : term -> term
  | TmAssign : term -> term -> term
  | TmLocation : label -> term
  | TmUnit : term
.

(** Make a module of [term]---we want it to implement [VALUE] because we store 
terms which are values in the store *)

Module Trm <: VALUE.
  Definition A := term.

  Definition equal (x y : term) := x = y.
 
  Lemma equal_refl : forall (t : term), equal t t.
  Proof.
  intros.
  unfold equal.
  reflexivity.
  Qed.
  Lemma equal_sym : forall (x y : term), equal x y -> equal y x.
  Proof.
  unfold equal.
  intros.
  symmetry.
  assumption.
  Qed.
  Lemma equal_trans : forall (x y z : term), equal x y -> equal y z -> equal x z.
  Proof.
  unfold equal.
  intros.
  subst x.
  subst z.
  reflexivity.
  Qed.

End Trm.  

(** The predicate [isval] is true if our term is a value. Only values are 
allowed in the store. *) 
Inductive isval : term -> Prop :=
  | VUnit : isval TmUnit
  | VLocation : forall (l : label), isval (TmLocation l)
.

(** [Val] combines a [term] with a proof that the term is a value. *) 
Module Val <: VALUE with Definition A := sig isval.
  Definition A := sig isval.
  Definition equal (x y : A) :=  
    match x,y with
     | exist  x _, exist  y _ => Trm.equal x y
    end
  .
  Lemma equal_refl : forall (x : A), equal x x.
  Proof.
  unfold equal.
  destruct x.
  apply Trm.equal_refl.
  Qed.

  Lemma equal_sym : forall (x y : A), equal x y -> equal y x.
  Proof.
  unfold equal.
  destruct x; destruct y.
  intros.
  apply Trm.equal_sym.
  assumption.
  Qed.

  Lemma equal_trans : forall (x y z : A), equal x y -> equal y z -> equal x z.
  Proof.
  unfold equal.
  destruct x; destruct y; destruct z.
  intros.
  eapply Trm.equal_trans.
  apply H.
  apply H0.
  Qed. 

End Val.  

(** The [Store] maps labels to terms that are values. *)

Module ValOpt := Option(Val).
Module Store := Map_Wf_Ext(ValOpt).

(** The data structure that models the store. *) 
Definition store := Store.store.

(** The data structure that models the store typing. *)
Definition store_typing := StoreTyping.store.

(** Type Schemas *)

Inductive typeof : (term * store_typing) -> ty -> Prop :=
  | TLoc : forall (t : ty) (s : store_typing) (l : label),
        StoreTyping.Opt.equal (StoreTyping.lookup s l) (Some t) ->
           typeof (TmLocation l, s) (TyRef t)
  | TRef : forall (tm : term) (t : ty) (s : store_typing),
        typeof (tm, s) t ->
           typeof (TmRef tm, s) (TyRef t)     
  | TDeref : forall (tm : term) (t : ty) (s : store_typing),
        typeof (tm, s) (TyRef t) ->
           typeof (TmDeref tm, s) t
  | TAssign : forall (tm1 tm2 : term) (t : ty) (s : store_typing),
        typeof (tm1,s) (TyRef t) ->
           typeof (tm2, s) t ->
              typeof (TmAssign tm1 tm2, s) TyUnit
  | TUnit : forall (s : store_typing), typeof (TmUnit, s) TyUnit
. 

(** A store is well\-typed with respect to a store typing if their domains
are equal and any term in the store is correctly typed by the store typing. *)

Inductive well_typed : store_typing -> store -> Prop :=
  | Well_typed : forall(st : store_typing) (s : store),
      StoreTyping.dom (st) = Store.dom (s) ->
        (forall (l : label) (t : ty) (tm : term) (H : isval tm), In l (Store.dom s) -> 
          StoreTyping.Opt.equal (StoreTyping.lookup st l) (Some t) -> 
            Store.Opt.equal (Store.lookup s l) (Some (exist isval tm H)) ->
               typeof (tm, st) t) -> 
                 well_typed st s.

(** The evaluation rules. *)

Inductive eval : (term * store) -> (term * store) -> Prop := 
  | ERefV : forall (t : term) (s : store) (H : isval t),
      let (l, s') := Store.allocate s (exist isval t H) in
           eval (TmRef t, s) (TmLocation l, s') 
  | ERef : forall (t t' : term) (s s' : store),
        eval (t,s) (t',s') ->
            eval (TmRef t, s) (TmRef t', s')
  | EDerefLoc : forall (t' : term) (H : isval t') (s : store) (l : label), 
          Store.Opt.equal (Store.lookup s l) (Some (exist isval t' H))  ->
            eval (TmDeref (TmLocation l), s) (t', s)
  | EDeref : forall (t t' : term) (s s' : store),
        eval (t,s) (t',s') ->
            eval (TmDeref t, s) (TmDeref t', s')
  | EAssign : forall (t2 : term) (s : store) (l : label) (H : isval t2),  
        eval (TmAssign (TmLocation l) t2, s) (TmUnit , Store.update s l (exist isval t2 H))
  | EAssign1 : forall (t1 t1' t2 : term) (s s' : store),
        eval (t1,s) (t1',s') ->
            eval (TmAssign t1 t2, s) (TmAssign t1' t2, s')
  | EAssign2 : forall (t1 t2 t2': term) (s s' : store),
        isval t1 ->      
           eval(t2,s) (t2',s') ->
               eval (TmAssign t1 t2, s) (TmAssign t1 t2', s')
.

(** The proof of progress. If the store is well typed with respect to the store
typing and a term can be typed then either the term is a value or it can take
a further step of evaluation. *)

Theorem progress : forall (st : store_typing) (s : store) (t : term),
  well_typed st s ->
     ex (fun x => typeof (t, st) x) ->
        {isval t} + {ex (fun x => eval (t, s) x)}. 
Proof.
  intros.
  induction t.
  right.
  destruct H0.
  inversion_clear H0.
  induction (IHt (ex_intro (fun x => typeof (t,st) x) t0 H1)).
  exists (let (l, s') := Store.allocate s (exist isval t a) in (TmLocation l, s')).
  apply (ERefV t s a).
  destruct b.
  destruct x0.
  exists (TmRef t1, s0).
  constructor.
  assumption.
  right.
  destruct H0.
  inversion_clear H0.
  induction (IHt (ex_intro (fun x => typeof (t, st) x) (TyRef x) H1)).
  destruct a.
  inversion_clear H1.
  inversion_clear H1.
  generalize 
    ((ex_intro (fun x => StoreTyping.Opt.equal (StoreTyping.lookup st l) (Some x)) x ) H0).
  intros.
  generalize ((proj1 (proj1 (StoreTyping.lookup_correct st l))) H1).
  intros.
  destruct H. 
  rewrite H in H2. 
  generalize ((proj2 (proj1 (Store.lookup_correct s l))) H2).
  intros.
  destruct H4.
  destruct x0.
  exists (x0, s).
  eapply EDerefLoc.
  apply H4.
  destruct b.
  destruct x0.
  exists (TmDeref t0, s0).
  constructor.
  assumption.
  right.
  destruct H0.
  inversion_clear H0.
  induction (IHt1 (ex_intro (fun x => typeof (t1, st) x) (TyRef t) H1)).
  induction (IHt2 (ex_intro (fun x => typeof (t2, st) x) t H2)).
  destruct a.
  inversion_clear H1.
  exists (TmUnit , Store.update s l (exist isval t2 a0)).
  constructor.
  destruct b.
  destruct x0.
  exists (TmAssign t1 t0, s0).
  constructor.
  assumption.
  assumption.
  destruct b.
  destruct x0.
  exists (TmAssign t0 t2, s0).
  constructor.
  assumption.
  left.
  constructor.
  left.
  constructor.
Defined.

(** This theorem relates the StoreTyping and the Store **)
Theorem preservation_aux3 : forall (st : store_typing) (s : store) (H : well_typed st s),
  forall n' : nat, In n' (StoreTyping.dom st) <-> Store.next s > n'.
Proof.
  intros.
  inversion_clear H.
  rewrite H0.
  destruct s.
  unfold Store.dom.
  simpl.
  apply next_gt.
Qed. 

(** If you update the store with a term of the same type as the term that
  is currently in the store at that location the store remains well typed *)
Lemma preservation_aux : forall (st : store_typing) (s : store) (l : label) (t : ty) (tm : term) (H : isval tm),
  well_typed st s ->
    StoreTyping.Opt.equal (StoreTyping.lookup st l) (Some t) -> 
       typeof (tm, st) t ->
          well_typed st (Store.update s l (exist isval tm H)).
Proof.
  intros.
  constructor.
  generalize (Store.update_preserves_dom s l (exist isval tm H)).
  intros.
  rewrite H3.
  inversion_clear H0.
  assumption.
  intros.
  induction (Peano_dec.eq_nat_dec l0 l).
  subst l0.
  rewrite (Store.update_preserves_dom s l (exist isval tm H)) in H4.
  generalize ((proj1 (proj1 (proj2 (Store.update_correct s l (exist isval tm H))))) H4).
  intros.
  generalize (Store.Opt.equal_sym _ _ H6). 
  intros. 
  generalize (Store.Opt.equal_trans _ _ _ H8 H7).
  intros.
  inversion_clear H9.
  inversion H10.
  subst tm0.
  generalize (StoreTyping.Opt.equal_sym _ _ H1).
  intros. 
  generalize (StoreTyping.Opt.equal_trans _ _ _ H9 H5). 
  intros.
  inversion_clear H11.
  inversion H12.
  subst t0.
  assumption.
  inversion_clear H0.
  eapply (H8 l0).
  rewrite (Store.update_preserves_dom s l (exist isval tm H)) in H4.
  assumption.
  assumption.
  generalize (((proj1 (Store.update_correct s l (exist isval tm H))) l0) b).
  intros.
  generalize (Store.Opt.equal_sym _ _ H0). 
  intros. 
  eapply (Store.Opt.equal_trans _ _ _ H9 H6).
Qed.

(** the type of a term is preserved in a store which is an extension of the
store in which the term is originally typed. *)
Lemma preservation_aux2 : forall (st st' : store_typing) (tm : term) (t : ty),
  typeof (tm,st) t -> 
     StoreTyping.extends st' st ->
       typeof(tm, st') t.
Proof.
  double induction tm t; intros; try solve [inversion H]; try solve [inversion H0; inversion H1; inversion H2].
  constructor.
  apply H0.
  inversion_clear H1.
  assumption.
  assumption.
  constructor.
  apply H0.
  inversion_clear H1.
  assumption.
  assumption.
  constructor.
  apply H.
  inversion_clear H0.
  assumption.
  assumption.
  inversion_clear H2.
  inversion_clear H1.
  eapply TAssign.
  apply H.
  apply H3.
  assumption.
  apply H0.
  assumption.
  assumption.
  inversion_clear H1.
  constructor.
  inversion_clear H0.
  generalize (ex_intro (fun x => StoreTyping.Opt.equal (StoreTyping.lookup st l) (Some x)) t0 H1).
  intros.
  generalize ((proj1 (proj1 (StoreTyping.lookup_correct st l))) H0).
  intros.
  generalize (H2 l H3).
  intros.
  generalize (StoreTyping.Opt.equal_sym _ _ H4).
  intros.
  apply (StoreTyping.Opt.equal_trans _ _ _ H5 H1).
  constructor.
Qed.


(** The case where an element gets allocated in the store is dealt with
separately. *)
Theorem preservation_allocation : forall (st : store_typing) (s : store) (tm : term) (t : ty) (H : isval tm),
  let (l, s') := Store.allocate s (exist _ tm H) in
  well_typed st s ->
    typeof (TmRef tm, st) t ->
      eval (TmRef tm, s) (TmLocation l, s') ->
        ex (fun x => StoreTyping.extends x st /\ typeof (TmLocation l, x) t /\ well_typed x s').
Proof.
  simpl.
  intros.
  inversion_clear H1.
  pose (StoreTyping.add st (Store.next s) t0 (preservation_aux3 st s H0)).
  exists s0.
  split.
  unfold s0.
  apply StoreTyping.add_extends.
  split.
  constructor.
  unfold s0.
  apply (proj2 (StoreTyping.add_correct st (Store.next s) t0 (preservation_aux3 st s H0))).
  constructor.
  unfold StoreTyping.dom.
  unfold Store.dom.
  simpl.
  inversion_clear H0.
  unfold StoreTyping.dom  in H1.
  unfold Store.dom in H1.
  rewrite H1.
  reflexivity.
  intros.
  unfold Store.lookup in H6.
  simpl in H6.
  induction (Peano_dec.eq_nat_dec (Store.next s) l).
  subst l.
  unfold s0 in H5.
  generalize (proj2 (StoreTyping.add_correct st (Store.next s) t0 (preservation_aux3 st s H0))).
  intros.
  generalize (StoreTyping.Opt.equal_sym _ _ H7).
  intros.
  generalize (StoreTyping.Opt.equal_trans _ _ _ H8 H5).
  intros.
  inversion_clear H9.
  inversion H10.
  unfold s0.
  subst t0.
  clear H10.
  clear H8.
  clear H7.
  clear H5.
  inversion_clear H6.
  inversion H5.
  subst tm0.
  clear H5.
  eapply preservation_aux2. 
  apply H3. 
  apply StoreTyping.add_extends.
  unfold s0 in H5.
  cut (l <> Store.next s).
  intros.
  generalize (proj1 ((proj1 (StoreTyping.add_correct st (Store.next s) t0 (preservation_aux3 st s H0))) l) H7).
  intros.
  generalize (StoreTyping.Opt.equal_trans _ _ _ H8 H5).
  intros.
  inversion_clear H0.
  unfold Store.dom in H4.
  simpl in H4.
  destruct H4.
  contradiction.
  generalize (H11 l t1 tm0 H1 H4 H9 H6).
  intros.
  eapply preservation_aux2.
  apply H12.
  unfold s0.
  apply StoreTyping.add_extends.
  auto.
Qed.
 
(** The actual preservation theorem. We demonstrate the existence of 
some store typing such that the type of the term after evaluation is the
type of the previous term. *)
Theorem preservation : forall (st : store_typing) (s s': store) (tm tm' : term) (t : ty),
  well_typed st s ->
    typeof (tm, st) t ->
      eval (tm, s) (tm', s') ->
        ex (fun x => StoreTyping.extends x st /\ typeof (tm',x) t /\ well_typed x s').
Proof.
  double induction tm t.
  (* The subterm of a reference is a value, so we allocate for the value *)
  intros.
  inversion H3; try (subst t1; subst s; subst tm'; subst s').
  apply (preservation_allocation st s0 _ (TyRef t) H4).
  assumption.
  assumption.
  apply H3. 
  (* The subterm of a reference is evaluated *)
  inversion_clear H2.
  destruct (H0 _ _  H1 H4 H5). 
  exists (x).
  split.
  tauto.
  split.  
  constructor.
  tauto.
  tauto.
  (* The reference is typed as unit, so we can eliminate *)
  intros.
  inversion H1.
  (* The dereference is typed as a reference *)
  intros.
  inversion H3; try (subst t1; subst s; subst tm'; subst s').
  (* The subterm of the dereference is a location, so the store typing does not change *)
  exists st. 
  split. 
  apply StoreTyping.extends_equal. 
  split.
  inversion_clear H1.
  eapply H7.  
  generalize (ex_intro (fun x => Store.Opt.equal (Store.lookup s0 l) (Some x)) _ H6). 
  intros. 
  eapply (proj1 (proj1 (Store.lookup_correct s0 l)) H1). 
  inversion_clear H2.
  inversion H1.
  assumption.
  apply H6.
  assumption.
  (* The subterm of the dereference can be evaluated further. *)
  inversion_clear H2.
  destruct (H0 t' (TyRef (TyRef t)) H1 H4 H5).
  exists x.
  split.
  tauto.
  split.
  constructor.
  tauto.
  tauto.
  (* The dereference is typed as a unit *)
  intros.
  inversion H2; try (subst t; subst s; subst tm'; subst s').
  (* Again, the subterm is a location *)
  exists st.
  split.
  apply StoreTyping.extends_equal.
  split.
  inversion_clear H0.
  eapply H6.
  generalize (ex_intro (fun x => Store.Opt.equal (Store.lookup s0 l) (Some x)) _ H5). 
  intros. 
  eapply (proj1 (proj1 (Store.lookup_correct s0 l)) H0). 
  inversion_clear H1.
  inversion H0.
  assumption.
  apply H5.
  assumption.
  (* Again, the subterm can be evaluated further *)
  inversion_clear H1.
  destruct (H t' (TyRef TyUnit) H0 H3 H4).
  exists x.
  split.
  tauto.
  split.
  constructor.
  tauto.
  tauto.
  (* The assignment is typed as a reference, so we eliminate *)
  intros.
  inversion_clear H3. 
  (* The assignment is typed as a unit *) 
  intros.
  inversion_clear H2.
  inversion H3; try (subst t; subst t0; subst s; subst tm'; subst s').
  (* The store is updated in this step, so store typing does not change *)
  exists st.
  split.
  apply StoreTyping.extends_equal.
  split.
  constructor.
  eapply preservation_aux.
  assumption.
  inversion H4.
  apply H7.
  assumption.
  (* The left hand side of the assignment must be reduced. *)
  destruct (H t1' (TyRef t2) H1 H4 H6). 
  exists x.
  split.
  tauto.
  split.
  eapply TAssign.
  apply (proj1 (proj2 (H2))).
  eapply preservation_aux2.
  apply H5.
  tauto.
  tauto.
  (* The right hand side of the assignment must be reduced. *)
  destruct (H0 t2' t2 H1 H5 H11).
  exists x.
  split.
  tauto.
  split.
  eapply TAssign.
  eapply preservation_aux2.
  apply H4.
  tauto.
  tauto.
  tauto.
  (* A location cannot be evaluated; so this case is eliminated. *)
  intros.
  inversion H2.
  (* For variety, we eliminate this case because the location is typed as unit *)
  intros.
  inversion H0.
  (* The unit is typed as a reference; so this case is eliminated. *)
  intros.
  inversion H1.
  (* Unit cannot be evaluated; so this case is eliminated. *)
  intros.
  inversion H1.
Qed.
