(** Proof of type soundness for the lambda calculus with references. 
Proof based on the development in Pierce, "Types and Programming Languages" *)

(** copyright Mulhern *)

Require Import List.
Require Gt.
Require Le.
Require Peano_dec.
Require Compare_dec.

Require Import util.
Require Import map.

(** [MAP_WF_EXT] is the type of a map that is well formed and can be
extended. *)
Module Type MAP_WF_EXT.
  Declare Module Opt : OPTION.
  Definition V := Opt.V.A.
  Parameter store : Set.
  Parameter next : store -> nat.
  Parameter dom : store -> list nat.
  Parameter lookup : store -> nat -> Opt.A.
  Parameter lookup_correct : forall (s : store) (n : nat),
  ((ex (fun b => Opt.equal (lookup s n) (Some b))) <-> In n (dom s)) /\
  (Opt.equal (lookup s n) None <-> ~ In n (dom s)).
  Inductive extends : store -> store -> Prop :=
    | Extends : forall (s s' : store),
      (forall (n : nat), In n (dom s) -> Opt.equal (lookup s n) (lookup s' n)) ->
      extends s' s.
  Parameter update : store -> nat -> V -> store.
  Parameter update_preserves_dom : forall (m : store) (n : nat) (newb : V),
    dom(update m n newb) = dom m.
  Parameter update_correct : forall (s : store) (n : nat) (newB : V),
     (forall (n' : nat), n' <> n -> Opt.equal (lookup (update s n newB) n') (lookup s n')) /\
     (In n (dom s) <-> Opt.equal (lookup (update s n newB) n) (Some newB)) /\
     (~ In n (dom s) <-> Opt.equal (lookup (update s n newB) n) None). 
  Parameter allocate : store -> V -> (nat * store).
  Parameter allocate_correct : forall (s : store) (v : V),
    (forall (z : nat), 
      z <> (next s) <-> 
        Opt.equal (lookup s z) (lookup (snd (allocate s v)) z)) /\
     Opt.equal (lookup (snd (allocate s v)) (next s)) (Some v).
  Parameter allocate_increments_dom : forall (s : store) (v : V),
     dom (let (l, s') := allocate s v in s') = (next s)::(dom s).
  Parameter add : forall (s : store) (n : nat), V -> (forall (n' : nat), In n' (dom s) <-> n > n') -> store.
  Parameter add_correct : forall (s : store) (n : nat) (v : V)
  (H : forall (n' : nat), In n' (dom s) <-> n > n'),
    (forall (n' : nat), n' <> n <-> Opt.equal (lookup s n') (lookup (add s n v H) n')) /\
      Opt.equal (lookup (add s n v H) n) (Some v). 
  Parameter add_extends : forall (s : store) (n : nat) (v : V) 
    (H : forall (n' : nat), In n' (dom s) <-> n > n'),
      extends (add s n v H) s.
  Parameter add_increments_dom : forall (s : store) (n : nat) (v : V)
    (H : forall (n' : nat), In n' (dom s) <-> n > n'),
      dom (add s n v H) = n::dom s.
End MAP_WF_EXT.

(** [Map_Wf_Ext] is an implementation of [MAP_WF_EXT] *)
Module Map_Wf_Ext (Val : OPTION) <: MAP_WF_EXT.

  Module Opt := Val.

  Definition V := Opt.V.A.
	 
  (** We want to use the Map module. *)
  Module Internal := Map(Val).

  (** a store is a record containing
    a map,
    a proof of well formedness of the map,
    the next index to use,
    a proof that the next index is larger than the indices in the map *)
  Record store_int : Set := mkStore {
    internal : Internal.map;
    internal_wf_gt : Internal.well_formed_gt internal;
    next : nat;
    next_gt : (forall (n : nat), In n (Internal.dom internal) <-> next > n)
  }.

  Definition store := store_int.

  Definition dom (s : store) : list nat := Internal.dom (s.(internal)).

  (** We relate the definition of [dom] to the internal structure of the
  record. *)
  Lemma next_gt_ext : forall (s : store) (n : nat), 
    In n (dom s) <-> s.(next) > n.
  Proof.
    destruct s.
    unfold dom.
    simpl.
    assumption.
  Qed.

  (** Two stores are equal if their maps are equal. *)
  Inductive equal : store -> store -> Prop :=
    | Equal : forall (s s' : Internal.map) (n n' : nat) 
                (I : Internal.well_formed_gt s) 
                (I' : Internal.well_formed_gt s')
                (P : forall (z : nat), In z (Internal.dom s) <-> n > z) 
                (P' : forall (z : nat), In z (Internal.dom s') <-> n' > z), 
       Internal.equal s s' -> equal (mkStore s I n P) (mkStore s' I' n' P')
  .

  (** Proofs of equivalence properties. *)
  Lemma equal_refl : forall (x : store), equal x x.
  Proof.
    destruct x.
    constructor.
    apply Internal.equal_refl.
  Qed.

  Lemma equal_sym : forall (x y : store), equal x y -> equal y x.
  Proof.
    destruct x; destruct y.
    intros.
    constructor.
    apply Internal.equal_sym.
    inversion_clear H.
    assumption.
  Qed.

  Lemma equal_trans : forall (x y z : store), equal x y -> equal y z -> equal x z.
  Proof.
    destruct x; destruct y; destruct z.
    intros.
    constructor.
    inversion_clear H.
    inversion_clear H0.
    eapply Internal.equal_trans.
    apply H1.
    apply H.
  Qed.

  (** [lookup] is defined in terms of its internal representation. *)
  Definition lookup (s : store) n : option V := Internal.lookup (s.(internal)) n.

  (** correctness is defined in terms of the correctness of the internal representation *)
  Lemma lookup_correct : forall (s : store) (n : nat),
    ((ex (fun b => Opt.equal (lookup s n) (Some b))) <-> In n (dom s)) /\
    (Opt.equal (lookup s n) None <-> ~ In n (dom s)).
  Proof.
    destruct s.
    intros.
    apply (Internal.lookup_correct internal0 n).
  Qed.

  (** Equivalence is defined in terms of the equivalence of the internal representation *)
  Lemma lookup_equivalent : forall s s' : store,
    equal s s' -> forall n : nat, Opt.equal (lookup s n) (lookup s' n).
  Proof.
    destruct s; destruct s'.
    intros.
    unfold lookup.
    simpl.
    apply Internal.lookup_equivalent.
    inversion_clear H.
    assumption.
  Qed.

  (** s' extends s *)
  Inductive extends : store -> store -> Prop :=
    | Extends : forall (s s' : store),
      (forall (n : nat), In n (dom s) -> Opt.equal (lookup s n) (lookup s' n)) ->
      extends s' s.
  
  (** If the stores are equal the extends relationship must hold *)
  Lemma extends_equal : forall (s : store), extends s s.
  Proof.
    intros.
    constructor.
    intros.
    apply Opt.equal_refl.
  Qed.

  (** After an internal update [next] is still greater than all keys in the domain *)
  Lemma update_aux : forall (s : store) (n : nat) (b : V),
    forall  (n0 : nat),
      In n0 (Internal.dom (Internal.update (internal s) n b)) <-> next s > n0.
  Proof. 
    destruct s.
    simpl.
    intros.
    rewrite (Internal.update_preserves_dom internal0 n b).
    apply next_gt0.
  Qed.
   
  (** external definition of [update] *)
  Definition update s n b := 
    mkStore 
      (Internal.update s.(internal) n b) 
      (Internal.update_preserves_wf_gt  s.(internal) n b s.(internal_wf_gt))
      s.(next) 
      (update_aux s n b)
  .

  (** using internal map representation to show update preserves dom *)
  Lemma update_preserves_dom : forall (m : store) (n : nat) (newb : V),
    dom(update m n newb) = dom m.
  Proof.
    destruct m. 
    unfold update.
    unfold dom.
    simpl.
    intros.
    rewrite (Internal.update_preserves_dom internal0 n newb).
    reflexivity.
  Qed.

  Lemma update_preserves_store : forall (m : store) (n : nat) (newb :V),
      ~In n (dom m) -> equal m (update m n newb).
  Proof.
    intros.
    destruct m.
    unfold update.
    unfold dom.
    unfold internal.
    simpl.
    intros.
    constructor.
    apply (Internal.update_preserves_map internal0 n newb H).
  Qed.
  
  Lemma update_correct : forall (s : store) (n : nat) (newB : V),
     (forall (n' : nat), n' <> n -> Opt.equal (lookup (update s n newB) n') (lookup s n')) /\
     (In n (dom s) <-> Opt.equal (lookup (update s n newB) n) (Some newB)) /\
     (~ In n (dom s) <-> Opt.equal (lookup (update s n newB) n) None). 
  Proof.
    intros.
    apply (Internal.update_correct s.(internal) n newB).
  Qed.
  
  (** When a new element is added to the store, a new proof that the 
    updated next element is larger than the elements in the map must be
    supplied. This lemma supplies that proof. *)
  Lemma allocate_aux : forall (m : Internal.map) (next : nat) (v : V) (W : Internal.well_formed_gt m) (P : forall (n : nat), In n (Internal.dom m) <-> next > n),
    forall (z : nat), In z (Internal.dom ((next, v)::m))  <-> (S next) > z.
  Proof.
    simpl.
    intros.
    split.
    intros.
    destruct H.
    subst z.
    omega.
    generalize ((proj1 (P z)) H). 
    omega. 
    intros. 
    induction (Peano_dec.eq_nat_dec next0 z). 
    subst z.
    tauto.
    right.
    apply (proj2 (P z)).
    omega. 
  Qed.           
  
  (** [allocate] returns the index to the just allocated value and the
    updated store. *)
  Definition allocate (s : store) (v : V) :=
  (s.(next),
  mkStore
    ((s.(next), v)::s.(internal)) 
    (Internal.ConsWellFormedGt
      s.(internal)
      s.(next)
      v
      s.(internal_wf_gt)
      s.(next_gt)
    )
    (S s.(next)) 
    (allocate_aux s.(internal) s.(next) v s.(internal_wf_gt) s.(next_gt)))
  .  

  (** [lookup] for any number other than next will give the same answer
   after the allocation step as before, and for next the value returned by
   [lookup] equals the value added by [allocate] *)
  Lemma allocate_correct : forall (s : store) (v : V),
  (forall (z : nat), 
    z <> s.(next) <-> 
      Opt.equal (lookup s z) (lookup (snd (allocate s v)) z)) /\
      Opt.equal (lookup (snd (allocate s v)) s.(next)) (Some v).
  Proof.
    unfold lookup.
    unfold allocate.
    destruct s.
    simpl.
    intros. 
    split. 
    intros. 
    induction (Peano_dec.eq_nat_dec next0 z). 
    subst z.
    split.
    intros.
    tauto.
    intros.
    generalize (ex_intro (fun x : Internal.Opt.V.A => Internal.Opt.equal (Internal.lookup internal0 next0) (Some x)) v H).
    intros.
    generalize (proj1 (proj1 (Internal.lookup_correct internal0 next0)) H0).
    intros.
    generalize (proj1 (next_gt0 next0) H1).
    omega.
    split.
    intros.
    apply Opt.equal_refl.
    auto.
    induction (Peano_dec.eq_nat_dec next0 next0). 
    apply Opt.equal_refl. 
    tauto. 
  Qed. 

  (** After allocation the internal map structure is still well formed *)
  Lemma allocate_preserves_wf_gt : forall (s : store) (v : V),
    Internal.well_formed_gt s.(internal) ->
      Internal.well_formed_gt (snd (allocate s v)).(internal).
  Proof.
    destruct s.
    simpl.
    intros.
    constructor.
    assumption.
    assumption.
  Qed.

  (** After allocation the domain has one more element *)
  Lemma allocate_increments_dom : forall (s : store) (v : V),
    dom (let (l, s') := allocate s v in s') = s.(next)::(dom s).
  Proof.
    unfold dom.
    simpl.
    auto.
  Qed.

  (** [add] is like [allocate] except that the index value is not returned 
  and a proof that n is greater than all elements in the domain must be
  supplied. *)
  Definition add (s : store) (n : nat) (v : V) 
    (H : forall (n' : nat), In n' (dom s) <-> n > n') := mkStore
    ((n, v)::s.(internal)) 
    (Internal.ConsWellFormedGt
      s.(internal)
      n
      v
      s.(internal_wf_gt)
      H
    )
    (S n) 
    (allocate_aux s.(internal) n v s.(internal_wf_gt) H)
  . 
  
  (** We require [add] to be correct in the same way that [allocate] is *)
  Lemma add_correct : forall (s : store) (n : nat) (v : V)
  (H : forall (n' : nat), In n' (dom s) <-> n > n'),
    (forall (n' : nat), n' <> n <-> Opt.equal (lookup s n') (lookup (add s n v H) n')) /\
      Opt.equal (lookup (add s n v H) n) (Some v).
  Proof. 
    destruct s.
    unfold add.
    unfold lookup.
    unfold dom.
    simpl.
    split.
    intros.
    induction (Peano_dec.eq_nat_dec n n').
    split.
    intros.
    symmetry in a.
    contradiction.
    subst n'.
    intros.
    generalize (ex_intro (fun x : Internal.Opt.V.A => Internal.Opt.equal (Internal.lookup internal0 n) (Some x)) v H0).
    intros.
    generalize (proj1 (proj1 (Internal.lookup_correct internal0 n)) H1).
    intros.
    generalize (proj1 (H n) H2).
    intro.
    omega.
    split.
    intros.
    apply Opt.equal_refl.
    auto.
    induction (Peano_dec.eq_nat_dec n n).
    apply Opt.equal_refl.
    tauto.
  Qed.

  (** when we call [add] the new store extends the old store in a formal sense *)
  Lemma add_extends : forall (s : store) (n : nat) (v : V) 
    (H : forall (n' : nat), In n' (dom s) <-> n > n'),
      extends (add s n v H) s.
  Proof.
    intros.
    constructor.
    intros.
    generalize ((proj1 (H n0)) H0).
    intros.
    apply (proj1 ((proj1 (add_correct s n v H)) n0)).
    omega.
  Qed.

  Lemma add_increments_dom : forall (s : store) (n : nat) (v : V)
  (H : forall (n' : nat), In n' (dom s) <-> n > n'),
    dom (add s n v H) = n::dom s.
  Proof. 
    unfold dom.
    simpl.
    auto.
  Qed.

End Map_Wf_Ext.
