(** Proof of type soundness for the lambda calculus with references. 
Proof based on the development in Pierce, "Types and Programming Languages" *)

(** copyright Mulhern *)

Require Peano_dec.

(** A [VALUE] is just a wrapper for some Set type. Since we don't care
whether equality is decidable for values we leave it out. *)
Module Type VALUE.
  Parameter A : Set.
  Parameter equal : A -> A -> Prop.
  Parameter equal_refl : forall x : A, equal x x. 
  Parameter equal_sym : forall x y : A, equal x y -> equal y x.
  Parameter equal_trans : forall x y z : A, equal x y -> equal y z -> equal x z.
End VALUE.

(** An [OPTION] is just like the option type. However, since it is an option
on a [VALUE] we need to define equality in terms of the equality of the [VALUE]. 
As with [VALUE] we don't require decidability of equality.*)
Module Type OPTION.
  Declare Module V : VALUE.
  Definition A := option V.A.
  Inductive equal : A -> A -> Prop :=
    | EqualNone : equal None None
    | EqualSome : forall (a a' : V.A), V.equal a a' -> equal (Some a) (Some a')
  .
  Parameter equal_refl : forall x : A, equal x x. 
  Parameter equal_sym : forall x y : A, equal x y -> equal y x.
  Parameter equal_trans : forall x y z : A, equal x y -> equal y z -> equal x z.
End OPTION.

(** [Option] is an implementation of [OPTION] for any X that implements
[VALUE]. The proofs of reflexivity, symmetry, transitivity of equality are
just wrappers for the corresponding proofs on X. *)
Module Option (X : VALUE) <: OPTION.
 
  Module V := X.

  Export X.

  Definition A := option V.A.

  Inductive equal : A -> A -> Prop :=
    | EqualNone : equal None None
    | EqualSome : forall (a a' : V.A), V.equal a a' -> equal (Some a) (Some a')
  .

  Lemma equal_refl : forall (o : A), equal o o.
  Proof.
    destruct o.
    constructor.
    apply V.equal_refl.
    constructor.
  Qed.

  Derive Inversion_clear equal_some_inv 
    with (forall (a a' : V.A), equal (Some a) (Some a')) Sort Prop.

  Lemma equal_sym : forall (x y : A), equal x y -> equal y x.
  Proof.
    destruct x; destruct y; intros; try solve [inversion H].
    constructor.
    inversion H using equal_some_inv.
    apply V.equal_sym.
    assumption.
  Qed. 

  Lemma equal_trans : forall (x y z : A), equal x y -> equal y z -> equal x z.
  Proof.
    destruct x; destruct y; destruct z; intros; try solve [inversion H; inversion H0].
    constructor.
    inversion H0 using equal_some_inv.
    inversion H using equal_some_inv.
    apply V.equal_trans.
    assumption.
  Qed.

End Option.
